#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Functions for inferring data types from file.

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).
'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.0";

import string;

# symbol can be considered missing data
def is_missing(symbol):
	return symbol is None or symbol is type(None) or string.lower(str(symbol)) in [ '?', '', 'nan' ]; 

# If all known values are unique, a variable is considered an example ID
class identifier(str): 
	pass;

# Zero-variance variables.  Constants can be non-strings, but what's the difference?
class constant(str):	
	pass;

# is `x' a bona-fide real?
def isreal(x):
	try:
		float(x);
		return True;
	except:
		return False;

# does string `x' look enough like a real number?  
# (missing values could be real and are therefore real enough, 
# see:  is_missing)
def real_enough(x):
	return isreal(x) or is_missing(x);

# does string `x' look enough like an integer?  (missing values could be anything and are therefore real enough, see:  is_missing)
def int_enough(x):
	try:
		f = float(x);
		return int(f)==f;
	except ValueError:
		return is_missing(x);
	except TypeError:
		return is_missing(x);

# Infer type of a list of strings, which might be numerical
def infer(sample):
	if (all(map(int_enough,sample))):	# (holds true even for missing values)
		return int;
	elif (all(map(real_enough,sample))):
		return float;
	elif all_uniq(sample):
		return identifier;
	elif all_same(sample):
		return constant;
	else:
		return str;

# infer the type of each column, return a parallel list.
def infer_column_types(X):
	return map(infer, map(lambda column : map(lambda x:x[column],X), range(len(X[0]))));

# What is the type of the interable (skipping missing values)?
def current_type(V):
	for v in V:
		if not is_missing(v):
			return type(v)

# Column is zero-origin
def gencol(X, column):
	for x in X:
		yield x[column];

# what are the types of columns.  Not an inline, since if type(X[0][column]) is
# NoneType (missing data), we need to look further down the column.  note this
# is for X already in memory, not to be confused with metadata read from file.
def current_column_types(X):
	ans = []; # ans[column] -> type
	for column in range(len(X[0])): #assumes X has some examples
		ans += [current_type(gencol(X,column))]
	return ans;

# recast values in place, replace missing values with None (see:  is_missing)
def cast(X, T):
	for row in range(len(X)):	# for each example
		for column in range(len(T)):
			castf = T[column]; 
			if (castf==int):
				castf = lambda x : int(float(x)); # int fails to convert some strings (e.g. "0.") that float handles.
			X[row][column] = None if is_missing(X[row][column]) else castf(X[row][column]); 

# Are all values of X unique (i.e. no duplicates)
def all_uniq(X):
	S = sorted(X);
	return not any(map( lambda i : S[i-1]==S[i], range(1,len(S)) ));

# Are all values of X identical?
def all_same(X):
	it=iter(X);
	try:
		last = it.next();
		for x in it:
			if x!=last:
				return False;
	except StopIteration:
		pass;
	return True;

