/*

	2010/Feb, noto@cs.tufts.edu

	Jian Tang et al.
	"Enhancing Effectiveness of Outlier Detections for Low Density Patterns"
	PAKDD 2002.

	This is a modified version of COF which works on a TRAIN and TEST set
	(see comments for constants N,C, and C0).

	Input is comma-[and newline]-separated data.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

// distance(i,j) stored in `stored_dist' (in main).  Only allocated stored_dist where i > j 
#define dist(i,j) ((i > j) ? stored_dist[i][j] : stored_dist[j][i])

double distance(const double *x, const double *y, const int dim);
double ac_dist(const double *c, const int n);

int main(const int argc, const char **argv) {

	time_t t0 = time(NULL);
    
	FILE *verbose = stderr;     // set to NULL to turn off verbose progress messages.
	int vperiod;

	int i,j,p,q,k,f;

	if (argc != 7) {
		fprintf(stderr,"ERROR: expecting command line arguments: \n");
		fprintf(stderr,"k lower bound, k upper bound, no. training instances, no. testing instances, no. features, input file name\n");
		exit(-1);
	}

	const int T = atoi(argv[3]);		// training instances
	const int M = T + atoi(argv[4]);	// total instances
	const int F = atoi(argv[5]);		// num features

	int user_kLB = atoi(argv[1]); 
	if (user_kLB < 2) { fprintf(stderr,"Setting k lower bound <- 2 (%i is too small)\n", user_kLB); user_kLB=2; }
	int user_kUB = atoi(argv[2]); 
	if (user_kUB >= T) { fprintf(stderr,"Setting k upper bound <- %i (%i is too large)\n", T-1, user_kUB); user_kUB=T-1; }
	
	const int kLB = user_kLB;
	const int kUB = user_kUB;

	// The following change standard COF to the special version we use to
	// estimate COF over a train and test set.  The basic idea is that test
	// examples are "invisible" to everybody else so that each test can be
	// calculated independently.

	const int N = T;	// available neighbors (set to T to ignore test examples as neighbors)
	const int C = M-T;	// number of examples for which to calculate COF
	const int C0 = T;	// 1st example for which to calculate COF

	if (kLB <= 0) { fprintf(stderr,"ERROR: k lower bound must be greater than 0: %s\n", argv[1]); }
	if (M <= 0) { fprintf(stderr,"ERROR: M must be greater than 0: %s\n", argv[4]); }
	if (F <= 0) { fprintf(stderr,"ERROR: F must be greater than 0: %s\n", argv[5]); }

	// allocate space for data (examples x features)
	double **data = (double**)malloc(sizeof(double*) * M);	// freed after stored_dist is calculated
	for (i=0; i<M; i++) { data[i] = (double*)malloc(sizeof(double)*F); }

	// space for pairwise distance (need only lower triangle of training x
	// training instances, and distance from test instances to training
	// instances)
	double **stored_dist = (double**)malloc(sizeof(double*) * M); 	// freed after trails are calculated
	for (i=0; i<T; i++) { stored_dist[i] = (double*)malloc(sizeof(double)*i); } // must be i > j
	for (i=T; i<M; i++) { stored_dist[i] = (double*)malloc(sizeof(double)* (N < i ? N : i)); }
	
	// read data
	FILE *in = fopen(argv[6],"r");
	for (i=0; i<M; i++) { 
		for (f=0; f<F; f++) { 
			fscanf(in, "%lf,", data[i]+f); 
		}
	}
	fclose(in);

	vperiod = T/20 + 100 - (T/20)%100;  // adjust to a percentage of training set, round to nearest 100
	t0=time(NULL);

    if (verbose) { fprintf(verbose, "# Compute pairwise distance for %i training examples\n", T); }
	for (i=0; i<T; i++) {
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
		for (j=0; j<i; j++) { 
			stored_dist[i][j] = distance( data[i], data[j], F );
		}
	}

    if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Compute pairwise distance for %i training examples", difftime(time(NULL),t0), M-T); }
	t0 = time(NULL);
	for (i=T; i<M; i++) {
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
		for (j=0; j<N; j++) { 
			stored_dist[i][j] = distance( data[i], data[j], F ); 
		}	
	}

	// free data
	for (i=0; i<M; i++) { free(data[i]); }
	free(data);
   
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Find %i neighbors for each of %i examples", difftime(time(NULL),t0), kUB, M); }
	t0 = time(NULL);

	// find everybody's k (UB) nearest neighbors, build their trails, compute the ordered COST of those trails
	double **cost = (double**)malloc(sizeof(double*) * M);	// freed after max_COF is calculated
	int **path = (int**)malloc(sizeof(double*) * M);
	
	for (i=0; i<M; i++) {
		
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
	
		int *neighbors = (int*)malloc(sizeof(int) * kUB);	// freed at the bottom of the loop
		
		int met = 0; 	// neighbors found so far
		int adr = -1;	// index (in neighbors) of most distant neighbor

		for (j=0; j<N; j++) { 

			if (j==i) { continue; }

			if (met < kUB) { 
				if (met==0 || dist(i,j) >= dist(i, neighbors[adr])) { 
					adr = met;
				}
				neighbors[met++] = j; 
			} else if (dist(i,j) < dist(i,neighbors[adr])) { 
				// replace neighbor at adr with this guy (sorry, neighbor at adr, you're not a neighbor anymore)
				neighbors[ adr ] = j;

				// find new farthest neighbor
				adr = 0;
				for (k=1; k<met; k++) { 
					if (dist(i, neighbors[k]) > dist(i, neighbors[adr])) { 
						adr = k; 
					}
				}
			}

		}
		
		// neighbors now contains the k (UB) nearest neighbors, ties broken by smallest index

		// compute the trail/cost 
		path[i] = (int*)malloc(sizeof(double) * kUB);	//freed when max_COF is calculated
		cost[i] = (double*)malloc(sizeof(double) * kUB);

		int *bitmap = (int*)malloc(sizeof(int)*kUB); // bitmap[ * ] iff * is in the path, freed after loop
		for (k=0; k<kUB; k++) { bitmap[k]=0; }

		for (p=0; p<kUB; p++) {

			// find next nearest neighbor, path[p]
			cost[i][p] = 1e999;
			int best_k = -1;

			for (k=0; k<kUB; k++) { 
				// consider adding neighbors[k] to the path

				if (!bitmap[k]) { // not already in the path, 
					// compute distance from k to path

					double dist_k_path = dist(i, neighbors[k]);

					for (q=0; q<p; q++) { 
						double d = dist(path[i][q], neighbors[k]);
						dist_k_path = d < dist_k_path ? d : dist_k_path;
					}

					if (dist_k_path < cost[i][p]) {	
						// so this is the best distance to the path
						cost[i][p] = dist_k_path;
						path[i][p] = neighbors[k];
						best_k = k;
					}
				}
			}
			bitmap[ best_k ] = 1;

		}

		// now the trail (path,cost) is created.
		free(bitmap);
		free(neighbors);

	} // next trail

	for (i=0; i<M; i++) { free(stored_dist[i]); } free(stored_dist);
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Compute COF for %i <= k <= %i", difftime(time(NULL),t0), kLB, kUB); }
	t0 = time(NULL);

	double *max_COF = (double*)malloc(sizeof(double) * C);	// freed at end of main
	for (i=0; i<C; i++) { max_COF[i] = -1e999; }

	for (k=kLB; k<=kUB; k++) { 
    
		if (verbose) { fprintf(verbose, "."); }
		
		// compute the ac-dist for each example
		double *ac_dists = (double*)malloc(sizeof(double) * M);	// freed at bottom of loop
		for (i=0; i<M; i++) { 
			ac_dists[i] = ac_dist(cost[i], k+1); 
		}

		// for each test example,  compute ac-dist and average
		//	ac-dist for elements in its path

		for (i=C0; i<C0+C; i++) { 
		
			double sum_ac_dist_o = 0.0;
			for (p=0; p<k; p++) { 
				sum_ac_dist_o += ac_dists[ path[i][p] ];
			}
			
			double COF = k * ac_dists[ i ] 
			    / //  ---------------------
			             sum_ac_dist_o;

			// keep the largest COF
			max_COF[i-C0] = COF > max_COF[i-C0] ? COF : max_COF[i-C0];
		}
	
		free(ac_dists);

	}
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n", difftime(time(NULL),t0)); }

	for (i=0; i<M; i++) { 
		free(path[i]);
		free(cost[i]);
	}
	free(path);
	free(cost);

	// print the (maximum) COF for each test instance
	for (i=0; i<C; i++) { 
		fprintf(stdout, "%10.10g\n", max_COF[i]); 
	}

	free(max_COF);

	return 0;
} // main 

double distance(const double *x, const double *y, const int dim) { 
	int i;
	double ssd = 0.0;
	for (i=0; i<dim; i++) { 
		double diff = x[i] - y[i];
		ssd += diff * diff;
	}
	return sqrt(ssd);
}

double ac_dist(const double *dist_e, const int r) {
	int i;
	double sum = 0.0;
	for (i=1; i<r; i++) { 
		sum += (r - i) * dist_e[i-1];
	}
	return sum * 2 / (r * (r-1));
}	

