#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Utilities for learning with SVMs (using LIBSVM, SVM-Light file formats)

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.0";

import sys,os,string,time,tempfile,math,random;
from util import *;

ulimit_t = 604800; # module-wide training time limit

# Classifier models a function f( x ) -> y and can generate predictions 
class SVMClassifier:	# should match mf.Classifier interface
	def __init__(self,predict_cmd,svmfile,clean):
		(self.predict_cmd, self.svmfile, self.clean) = (predict_cmd, svmfile, clean);
	def __call__(self,x):
		[ prd ] = list(self.predictions([x])); 
		return prd;
	def predictions(self, X):
		for prd in predict(X, self.svmfile, self.predict_cmd):
			yield prd;
	def __str__(self):
		return "SVMClassifier(%s,%s)" % (self.predict_cmd, self.svmfile); 
	def __del__(self):
		if self.clean:
			##sys.stderr.write("# REMOVE svm model %s\n" % (self.svmfile)); 
			rm(self.svmfile);
			pass;
		else:
			##sys.stderr.write("# KEEP svm model %s\n" % (self.svmfile));
			pass;

# write X in libsvm-format to stream
# X:  list of (sparse vector,target)
def write(X, out):
	for (x,y) in X:
		out.write("%g\t%s\n" % (y, ' '.join(map(lambda i : "%i:%g" % (i+1, x[i]), range(len(x))))));

# tempfilename:  generate unique temporary file names
class TempfileGenerator:
	def __init__(self, dir):
		self.calls=0;
		self.dir = dir;
	def __call__(self,suffix="tmp",prefix="tmp"):
		self.calls+=1;
		ans = "%s/%s.%i-%i.%s" % (self.dir, prefix, os.getpid(), self.calls, suffix); 
		return ans;
tempfilename = TempfileGenerator(tempfile.gettempdir());	# module-wide directory

# Remove a temporary file (may be disabled for debugging)
def rm(filename):
	#sys.stderr.write("Warning:  Keeping temporary file %s\n" % (filename))
	os.unlink(filename);

# train a model 
# @param trainset set of (x,y) pairs
# @return model filename
def train(trainset, train_cmd, svmfile=None, out = NullStream()):
	trainfile = tempfilename("smx","trainset");
	svmfile = svmfile if svmfile is not None else tempfilename("svm","model");
	write(trainset,open(trainfile,'w')); 
	shell_cmd = "ulimit -t %i;  %s %s %s" % (ulimit_t, train_cmd, trainfile, svmfile);
	out.write("#%s: $ %s\n" % (time.strftime("%A %B %d %I:%M:%S %p %Y"), shell_cmd)); 
	map(lambda stream : finish(stream,out), proc(shell_cmd,sink=NullStream()));  
	if not os.path.exists(svmfile):
		raise Exception("Failed to learn model:  %s" % (shell_cmd)); 
	rm(trainfile);
	return svmfile;

# classify examples; generate predictions
def predict(testset, svmfile, cmd, out = NullStream()):
	testset = map(lambda x : (x,float()), testset); # add dummy targets for file format
	testfile = tempfilename("smx","testset");
	write(testset,open(testfile,'w'));
	outputfile = tempfilename("output", "testset");
	shell_cmd = "%s %s %s %s" % (cmd,testfile,svmfile,outputfile);
	out.write("#%s: $ %s\n" % (time.strftime("%A %B %d %I:%M:%S %p %Y"), shell_cmd)); 
	map(lambda stream : finish(stream,out), proc(shell_cmd,sink=NullStream()));  
	outputfile_stream = open(outputfile,'r');
	
	# normally this would be a generator, but we need to make sure the loop
	# completes and the files are deleted (this can happen because the user
	# knows how many results to expect)

	predictions = [];
	for line in outputfile_stream:
		record = map(lambda x : x.strip(), line.split())
		if record[0].lower()=='labels':	# special case:  LIBSVM (with -b 1 options) is printing a distribution, so return that.
			label_order = map(int,record[1:])
		elif len(record) > 1:
			p = []; 
			for i in range(len(label_order)):
				while label_order[i] >= len(p):
					p = p + [ 0.0 ];
				p[label_order[i]] = float(record[i+1]);
			predictions += [ p ];
		else:
			predictions += [ line.strip() ]	# map to float?
	
	outputfile_stream.close();
	if len(predictions) != len(testset):
		raise Exception("Number of predictions (%i) not equal to number of test set examples (%i):  %s" % (len(predictions),len(testset),shell_cmd));
	map(rm,[outputfile,testfile]);
	assert not filter(lambda p : p is None, predictions)	# no prediction can be 'None'
	return predictions;

# My gridsearch approach.  
#


# trainset:  set of (x,y) pairs where x are vectors, y are targets
# tcmd: train shell command, e.g. "svm_train"
# pcmd: predict shell command, e.g "svm_predict"
# search_parameters: a set of tuples (flag,start,(first,last,step)).  
# 	e.g. ('c', 0, (-7,3,1)) means search for -c between -7 and +3, 
#	starting at 0, and moving +1 each time (all in log space).
# base: log base
# metric: scoring metric maps a set of (obs,prd) pairs to some real.
# metric_cmp: compare scores, return negative iff first arg is better, 0 if they're the same
# metric_name: for the progress output
# metric_perfect:  scores this good or better immediately qualify.  it turns out that "perfect" accuracy scores are common enough to want to skip any additional processing
# folds:  for cross-validation
# folds_todo:  fold indecies to do.  Useful for holding aside 10% or 50% and doing one test only
# progress:  write progress here
# gibbs_loops:  set each variable this many times.
# fine_tuning_loops:  zoom in on the grid and research this many times.

def gridsearch( trainset, tcmd, pcmd, search_parameters, base, metric, metric_cmp, metric_name, metric_perfect, folds, folds_todo, progress = NullStream(), gibbs_loops = 1, fine_tuning_loops = 0):

	# log-scale search
	log = lambda x : -1e999 if x==0 else math.log(x) / math.log(base); # (base is param)
	exp = lambda lgv : base**lgv;

	# merge flags and values (solution is just a vector, need to refer back to componentflags to see what they mean)
	parameterize_solution = lambda v : dict(reduce( lambda ans,index : ans + [(componentflags[index],exp(v[index]))] , range(len(v)), [] ));		# create a mapping from flag to (linear scale) value
	gen_train_cmd = lambda v : "%s %s" % (tcmd, dict2parameters(parameterize_solution(v)));

	# define score of a vector
	pscore = lambda v : metric(gridsearch_crossval(trainset, folds, f=lambda A,B : gridsearch_eval(A,B,gen_train_cmd(v),pcmd)))

	# separate out parameters
	componentflags = []; 
	solution = [ ];	# current log values
	ranges = [ ];	# (logmin, logmax, logstep)
	for (flag,start,rangetuple) in search_parameters:	# search_parameters defined above
		componentflags += [ flag ];
		solution += [ start ];
		ranges += [ rangetuple ];
	
	# compare function caches results
	class Cmp:
		def __init__(self):
			self.cache={};
			self.best = None;
			self.calls = 0;
		def __call__(self, v1, v2):
			scores = [];	# scores will be [ score(v1), score(v2) ]
			for v in [v1,v2]:
				if tuple(v) in self.cache:
					scores += [ self.cache[tuple(v)] ];
				else:
					self.cache[tuple(v)] = pscore(v);
					scores += [ self.cache[tuple(v)] ];
					self.calls+=1;
					strv = ', '.join(reduce( lambda ans,index : ans + ["log(%s)=%7.3g" % (componentflags[index].replace('-',''), v[index])] , range(0,len(v)) , []));	
					progress.write("# %i\tv = <%s>\t%s = %s %s\n" % (self.calls, strv, metric_name, scores[-1], "**" if self.best is None or metric_cmp(scores[-1],self.best)<0 else ""));
					self.best = scores[-1] if self.best is None or metric_cmp(scores[-1],self.best) < 0 else self.best;
			return metric_cmp(scores[0], scores[1]);
		def perfect(self):
			return self.best is not None and metric_cmp(self.best,metric_perfect) <= 0	# can quit early if "perfect" score is achived 
	cmp = Cmp();

	# multiple Gibb's loops  
	cmp(solution,solution)	# this caches the initial default solution, which is only necessary for checking if it's perfect
	for gibbs_loop in range(gibbs_loops):
		for v in gibbs( solution, ranges, cmp ):	
			solution = v if (cmp(v,solution)<0) else solution;
	
	# (optional fine-tuning steps; could loop longer and narrow range as you search)
	for loops in range(fine_tuning_loops):
		# reset range for fine tuning
		step = 1.0 / 2.0**loops;	# zooom factor
		ranges = map( lambda v : (v-step,v+step,step/2.0) , solution ); 
		for v in gibbs( solution, ranges, cmp ):
			solution = v if (cmp(v,solution)<0) else solution;

	return parameterize_solution(solution)

# map a dict (e.g. {'c':34324.23423}, {'p':0.000123}) to a string (e.g.  "-c 34324.23423 -p 0.000123")
# see: gridsearch
dict2parameters = lambda D : ' '.join(map( lambda (flag,value) : "-%s %s" % (flag,value),  D.iteritems()));

# Evaluate train command tcmd on testset.  Generate (obs,prd) pairs.
def gridsearch_eval(trainset, testset, tcmd, pcmd):
	svmfile = train(trainset, tcmd); # no need to write process output 
	# break apart and recombine testset (x,y) so that I can make svm predictions en masse
	xit = map( lambda (x,y):x, testset );
	yit = iter(map( lambda (x,y):y, testset ));
	for prd in predict(xit, svmfile, pcmd):
		yield( yit.next() , prd );
	rm(svmfile);

# Generate results from generator f(trainset,testset) over cross-validaiton
def gridsearch_crossval(X, N, f, todo=None):
	X = list(X);
	for fold in (range(N) if todo is None else todo):
		## sys.stderr.write("# (gridsearch fold %s of %s:\n" % (fold,  range(N) if todo is None else todo))
		trainset = map(lambda i : X[i], filter(lambda i : i%N!=fold, range(len(X))));
		testset  = map(lambda i : X[i], filter(lambda i : i%N==fold, range(len(X))));
		for whatever in f(trainset,testset):
			yield whatever;

# Run a greedy sort of Gibbs sampling.  Try all values for all variables within
# range, extend range if necessary, stop when vector stops changing.
#
# v			variable setting
# ranges	min,max for each var (parallel to v)
# cmp		compare function cmp(a,b) returns neg,0,pos iff (a<b,a==b,a>b), lower is better--e.g. sort-by-best
#			(will call cmp.perfect(), to see if we can quit early--comment out calls if cmp is just a function)
def gibbs(v, ranges, cmp, search_order=None):

	if not search_order:
		search_order = range(len(v)); # default, all search_order from 0 to n-1.

	def test(v,i,alt_i):
		alt_v = v[:i] + [alt_i] + v[i+1:];
		return alt_v if cmp(v,alt_v) > 0 else v;

	v0 = v;
	while not cmp.perfect():
	
		for i in search_order: # for each component of the ones given, in order
		
			(first,last,step) = map(float,ranges[i]);

			alt_i = first;
			sign = lambda x : -1 if x < 0 else 1 if x > 0 else 0;
			while (first <= alt_i <= last) or (first >= alt_i >= last) or (sign(last - alt_i)==sign(step)):	# while in range or moving toward last
				v = test(v,i,alt_i); 
				if cmp.perfect():break;
				alt_i += step;

			while (v[i]==first):
				first = alt_i = first - step;
				v = test(v,i,alt_i); # going forward
				if cmp.perfect():break;
			while (v[i]==alt_i):	# while choice equals last option in range
				last = alt_i = last + step;
				v = test(v,i,alt_i); # going forward
				if cmp.perfect():break;

			ranges[i] = (first,last,step); # reset in case of any change

			yield v;

		if v==v0: # quit if no change
			break;
		else:
			v0=v;


'''
Some SVM-based learning functions.
'''

# Generate a random string
def junkey(length=10):
	return reduce(lambda a,x : a+x, random.sample("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ", length));

# classifier function learner
class Learner:
	def __init__(self, dir=tempfile.gettempdir(), clean=True, tcmd=None, pcmd=None, gridsearch=None, gridsearch_metric=None):
		self.calls = 0;
		self.dir, self.clean = dir, clean;
		self.tcmd, self.pcmd = tcmd, pcmd;
		self.gridsearch_parameters = gridsearch
		self.gridsearch_metric = gridsearch_metric
		self.hash = reduce(lambda h,c : h.replace(c,'.'), " /\\\t", tcmd)       # match the exact parameters only
		self.extension = "%s.model" % (os.path.basename(tcmd.split()[0])) # type of model matches calling program
	def __call__(self,trainset):
		self.calls += 1;
		svmfile = "%s/%s-%i.%s" % (self.dir, self.hash, self.calls, self.extension)
		if os.path.exists(svmfile) and os.path.getsize(svmfile):
			sys.stderr.write("# Use existing model:  %s\n" % (svmfile));
			pass;
		else:
			trainset = filter(lambda (x,y) : y is not None, trainset); # remove examples with missing targets
			if self.gridsearch_parameters:
				assert self.gridsearch_metric
				base = 10;
				log = lambda x : math.log(x)/math.log(base);
				parameters = gridsearch(trainset=trainset,tcmd=self.tcmd,pcmd=self.pcmd,\
					search_parameters = self.gridsearch_parameters,\
					base = base, \
					metric=self.gridsearch_metric,\
					metric_cmp = self.gridsearch_metric.cmp,\
					metric_name = str(self.gridsearch_metric),\
					metric_perfect = self.gridsearch_metric.perfect(),\
					folds=10, folds_todo=range(10), \
					progress = sys.stderr, \
					gibbs_loops = 1, fine_tuning_loops = 0);
			parameters = dict2parameters(parameters);
			train(trainset, "%s %s" % (self.tcmd,parameters), svmfile=svmfile)
		return SVMClassifier(self.pcmd, svmfile, clean=self.clean);


class MSE:
	def __init__(self):
		self.cmp = lambda a,b : -1 if a<b else 1 if a>b else 0
	def perfect(self):
		return 0.0;
	def __str__(self):
		return "MSE"
	def __call__(self,pairs):
		(ssd,N) = reduce(lambda (ssd,N),(o,p) : (ssd+((o-p)*(o-p)),N+1), map(lambda (o,p):(o,float(p)),pairs), (0.0,0));
		return (ssd/N) if N else 1e999; # ad-hack:  use infinity as a "bad result" for empty sets

class Accuracy:
	def __init__(self):
		self.cmp = lambda a,b : -1 if a>b else 1 if a<b else 0 # higher is better
	def perfect(self):
		return 1.0
	def __str__(self):
		return "Accuracy";
	def __call__(self,pairs):
		(correct,total) = reduce(lambda (c,t),(o,p) : (c+1,t+1) if (o<=0)==(p<=0) else (c,t+1), map(lambda (o,p):(o,float(p)),pairs), (0,0)); 
		return (float(correct) / float(total)) if total else 0.0; 

class HardAccuracy(Accuracy): # compatible only with the -b (use probabilities) option in LIBSVM
	def __str__(self):
		return "(Hard) Accuracy"
	def __call__(self,pairs):
		correct = total = 0;
		for (o,p) in pairs:	 # you are correct if the max probability prediction is the answer
			(maxprob,index) = sorted(map(lambda (index,prob) : (prob,index), iwalk(p)))[-1]	
			(total,correct)  =  (total+1, correct + 1 if index==o else correct)
		return float(correct) / total if total else 0.0;

class SoftAccuracy(Accuracy): # compatible only with the -b (use probabilities) option in LIBSVM
	def __str__(self):
		return "(Soft) Accuracy"
	def __call__(self,pairs):	# you are correct to the extent you assigned probability mass to the right answer
		(correct,total) = reduce(lambda (c,n),(o,p) : (c+float(p[int(o)]),n+1), pairs, (0.0,0))
		return correct / total if total else 0.0;

#
#import metric;
#class AUROCC:
#	def __init__(self):
#		self.cmp = lambda a,b : -1 if a>b else 1 if a<b else 0 # higher is better
#	def perfect(self):
#		return 1.0
#	def __str__(self):
#		return "AUROCC";
#	def __call__(self,pairs):
#		return metric.auc(metric.roc(pairs));
#

