#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Utilities for learning with Classification/Regression Trees (using WEKA)

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.0";

import sys,os,string,StringIO,time,tempfile;
from util import *;

wekapath = "weka"; # module-wide

# define tree algorithms
J48="J48"
REPTree="REPTree"
algorithms = [J48,REPTree]

# classifier function learner
# features:  list of (name, possible values (or just 'float' if it is a numerical type))
# The LAST feature is the class feature
class Learner:
	def __init__(self, features, algo, dir=tempfile.gettempdir(), keycode="frac-tree", clean=True):
		if algo not in algorithms:
			raise Exception("Unrecognized algorithm %s" % algo)
		self.algo, self.tmpdir, self.tmpkey, self.clean = algo, dir, keycode, clean
		self.calls = 0;
		# write ARFF format header to append to train/test files later
		out = StringIO.StringIO()
		out.write("%% ARFF File created %s\n" % (time.strftime("%Y %b %d %H:%M:%S")))
		out.write("@relation %s\n" % (self.tmpkey))
		for (name, values) in features:
			if values in [None,float,int]: # continuous feature
				out.write("@attribute %s real\n" % (name))
				self.tvals = None	# last feature is class feature, None indicates real-valued
			else:	# nominal
				out.write("@attribute %s {%s}\n" % (name, ','.join(gmap(str,values))))
				self.tvals = values;	# remember target features b/c tree prints them out weird
		out.write("@data\n");
		self.header = out.getvalue()
		out.close()
	# trainset -> classifier function
	def __call__(self,trainset):
		self.calls += 1;
		treefile = "%s/%s-%i.weka.%s.model" % (self.tmpdir, self.tmpkey, self.calls, self.algo)
		if os.path.exists(treefile):
			sys.stderr.write("# Use existing model %s\n" % (treefile)) 
			pass
		else:
			trainfile = "%s/%s-%i.train.arff" % (self.tmpdir, self.tmpkey, self.calls)
			write(trainfile, self.header, trainset)	# okay to write examples with missing target values, WEKA will just ignore them
			
			# Set pruning options (trees sometimes crash even with default parameters, so we have to make sure the options are valid)
			# REPTree does pruning by default (-P to turn it off; -N option means number of folds--must be <= number of examples)
			# J48 does not do pruning by default, requires -R option (and then -N is the number of folds)
			options = " -no-cv " # we're writing our models to file, so this turns off doing cross-validation to estimate accuracy and write it to the screen
			M = glen(gfilter(lambda (x,y):y is not None, trainset)); # number of valid training examples
			if M > 3:	
				options += " -R -N 3 " if self.algo in [J48] else " -N 3 "; # 3 is the default.
			else:	# good luck
				options += " -P " if self.algo in [REPTree] else ""	# 2 (or 1) examples; turn pruning off (we just want the options to be compatible with the input so the program doesn't crash)
			train(trainfile, treefile, self.algo, options);	# write treefile
			os.unlink(trainfile) 
		return Predictor(self.algo, self.header, treefile, self.clean, self.tvals, self.tmpdir, self.tmpkey)

class Predictor:
	def __init__(self,algo,header,treefile,clean,tvals,tmpdir,tmpkey):
		self.algo,self.header,self.treefile,self.clean,self.tvals,self.tmpdir,self.tmpkey = algo,header,treefile,clean,tvals,tmpdir,tmpkey
		self.calls=0
	def __call__(self,x):
		[ prd ] = list(self.predictions([x])); 
		return prd;
	def predictions(self, X): # Must maintain order of 'X'
		self.calls += 1
		testfile = "%s/%s-%i.test.arff" % (self.tmpdir, self.tmpkey, self.calls)
		write(testfile,self.header,gmap(lambda x : (x, self.tvals[0] if self.tvals else 0.0), X)); # add "dummy" targets to test file (since WEKA insists they be there and be valid values)
		predictions = list( predict(testfile, self.treefile, self.algo, options="") ) # read into memory so I can be sure to close the test file
		os.unlink(testfile);
		for prd in predictions:
			if self.tvals:# nominal, extract 1-origin feature number, map back to value
				prd_feature = self.tvals[ int(prd.split(':')[0])-1 ]
				# (double check conversion) # sys.stderr.write("# tree.Predictor.predictions:  convert prediction '%s' to '%s'\n" % (prd, prd_feature)) 
				assert prd_feature is not None;
				yield prd_feature;
			else:#real
				yield float(prd)
	def __del__(self):
		if self.clean:
			os.unlink(self.treefile)

# sub for learner and predictor
def write(outfile,header,xypairs):	
	out = open(outfile,'w')
	out.write(header)
	for row in map(lambda (x,y) : x + [y], xypairs):
		out.write("%s\n" % (','.join(map(lambda val : '?' if val is None else str(val), row))));
	out.close()

# create treefile model
def train(trainfile, treefile, algo, options):
	cmd = "java -classpath \"%s/weka.jar\" weka.classifiers.trees.%s %s -t \"%s\" -d \"%s\"" % (wekapath,algo,options,trainfile,treefile)
	# (echo cmd) # sys.stderr.write("# $ %s\n" % (cmd))
	(out,err) = proc(cmd); 
	output = out.read() if out else None;
	if out:  out.close();
	error = err.read() if err else None;
	if err:  err.close();
	if not os.path.exists(treefile):
		raise Exception("Failed to execute WEKA tree learner.\n\nCommand:\n\n%s\n\nError:\n\n%s\n\nOutput:\n\n%s\n\n" % (cmd, error, output))

# generate (string) predictions
def predict(testfile, treefile, algo, options):
	cmd = "java -classpath \"%s/weka.jar\" weka.classifiers.trees.%s %s -T \"%s\" -l \"%s\" -p 0" % (wekapath,algo,options,testfile,treefile)
	# (echo cmd) # sys.stderr.write("# $ %s\n" % (cmd))
	(out,err) = proc(cmd);
	if err:
		sys.stderr.write(err.read());
		raise Exception("Failed to execute WEKA tree learner:  %s" % (cmd)); 
	else:
		# read predictions into memory so I can be sure to close the file
		predictions = map(lambda line : line.split()[2], gfilter(lambda line: line and '0' <= line[0] <= '9', gmap(lambda line : line.strip(),out)))
		out.close()	
		for g in predictions:
			yield g;
