#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Supplement functions to train each feature predictor, compute normalized surprisal, etc.

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).
'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.0";

import sys,string,math,tempfile,time,os,os.path
import frac
from frac import svm,tree,parrot 
import format,stats,multinomial,autotype,sog,pdf
from util import *

# make all predictions 
def predict(A, B, targets=None, folds=10, cache=tempfile.gettempdir(), \
				LINEAR_SVM_MODEL=False, RBF_SVM_MODEL=False, TREE_MODEL=False):

	assert A and B;
	A=list(A)
	B=list(B)
	N = len(A[0]); # N = number of features
	
	targets = range(1,N+1) if targets is None else targets

	for target in targets:

		if not (1 <= target <= N):
			sys.stderr.write("Warning:  Feature #%i is out-of-bounds [1,%i].\n" % (target,N)); 
			continue;
		
		cachesub = chkdir("%s/f%i" % (cache,target));
		
		k = target - 1; # zero origin
		
		training_target_values = list(uniq(sorted(filter(lambda value:value is not None, map(lambda x : x[k],A)))))	# all distinct value of the training set
		if not training_target_values:
			sys.stderr.write("# Warning:  We will not train on feature %i, which zero valid values in the training set.\n" % (target))
			continue; 

		predictors = filter( lambda j : j != k , range(N))
		lset = frac.LSet(k,predictors);   # create a feature divider, data set -> (predictors,target) pairs of type frac.LSet
		target_type = autotype.current_type( gmap(lambda x : x[k], gcat((A,B))) );

		if len(training_target_values)==1:

			sys.stderr.write("# Note:  Trainset has only one possible value for feature %s (%s), so we use a \"parrot\" predictor.\n" % (target,training_target_values[0]))
			sublearn = parrot.Learner(training_target_values[0])
			(trainset, testset) = ( map(lset,A) , map(lset,B) )
			f = sublearn( filter(lambda (x,y) : y is not None, trainset )); # execute	
			
			# make training/testing set predictions
			yield ( target, target_type, \
					gmap(lambda (y,g):(y,g), frac.trainset_predict(trainset, sublearn, N=folds)) ,\
					gpairs(gmap(lambda (x,y) : y, testset), f.predictions(map(lambda (x,y) : x, testset))) )

		else:	

			# SVM models
			for ( _, settings) in filter(lambda (use_model,_) : use_model, [ \
				(RBF_SVM_MODEL, ( \
					( "svm-train -s 0 -t 2 -d 3 -e 0.1", "svm-predict", \
					  [ ('g', -1, (-7,-1,1)), ('c', 0, (-4, +2, 1)) ], frac.svm.Accuracy() ), \
					( "svm-train -s 3 -t 2 -d 3 -e 0.1", "svm-predict", \
					  [ ('g', -1, (-7,-1,1)), ('c', 0, (-4, +2, 1)), ('p',-1e999,(-7,-1,1)) ], frac.svm.MSE() ))), \
				(LINEAR_SVM_MODEL, ( \
					( "liblinear-train -s 4 -e 0.1", "liblinear-predict", \
					  [ ('c', 0, (-4, +2, 1)) ], frac.svm.Accuracy() ), \
					( "liblinear-reg-train -s 5 -e 0.1", "liblinear-reg-predict", \
					  [ ('c', 0, (-4, +2, 1)), ('p',-1e999,(-7,-1,1)) ], frac.svm.MSE() ))), \
				]):

				# For SVMs, target values must be integers, so translate string
				# labels to 0,1,2,3...  and define interpret to recast those
				# integers back to the original labels.
			
				(trainset, testset) = svm_reformat( map(lset,A) , map(lset,B) ); # change types to numbers (e.g. for svm)

				if target_type in [float,int]:	# continuous target
					( _ , (tcmd, pcmd, gridsearch, gridsearch_metric)) = settings # get the regression settings, which are second
					##sys.stderr.write("# (regression) settings:  tcmd=%s, pcmd=%s\n" % (tcmd,pcmd))
					interpret = lambda y:None if y is None else float(y); 
					sublearn = frac.svm.Learner( tcmd=tcmd, pcmd=pcmd, gridsearch=gridsearch, gridsearch_metric=gridsearch_metric, dir=cachesub, clean=False )
				elif target_type in [str]:	# nominal target
					( (tcmd, pcmd, gridsearch, gridsearch_metric), _ ) = settings # get the nominal settings
					##sys.stderr.write("# (classification) settings:  tcmd=%s, pcmd=%s\n" % (tcmd,pcmd))
					labels = sorted(uniq(filter(lambda y:y is not None,map(lambda (x,y):y, gcat((trainset,testset)))))) # mapping from 0-origin integers to feature values
					interpret = lambda i : None if i is None else labels[int(i)]
					trainset = map(lambda (x,y) : (x, None if y is None else labels.index(y)), trainset) # replace symbols with integers in input (leave y==None in--predictor will distcard them when training--so all trainset examples are accounted for)
					sublearn = frac.svm.Learner( tcmd=tcmd, pcmd=pcmd, gridsearch=gridsearch, gridsearch_metric=gridsearch_metric, dir=cachesub, clean=False )
				else:	
					raise Exception("Feature %i unknown type %s" % (target,target_type))

				# make train/testset predictions
				sys.stderr.write("# Learn SVM (%s) predictor for feature %i of %i (%s)\n" % (tcmd, target, N, time.strftime("%Y %b %d %H:%M:%S")))
				f = sublearn( filter(lambda (x,y) : y is not None, trainset ));
				yield ( target, target_type, \
						gmap(lambda (y,g):(interpret(y),interpret(g)), frac.trainset_predict(trainset, sublearn, N=folds)) ,\
						gpairs(gmap(lambda (x,y) : y, testset), gmap(interpret, f.predictions(map(lambda (x,y) : x, testset)))) )
				# Note:  need to interpret(y) in the *trainset* above b/c we changed all those targets

			if TREE_MODEL:  # (Also) Using Tree Model
				
				gcol = lambda column : gmap(lambda x : x[column], gcat((A,B))) # gcol: map column index to column generator
				flist = predictors + [k]; # feaature list, with target last
				xtypes = map(lambda column : autotype.current_type(gcol(column)),  flist) # feature/class types
				xvalues = map(lambda (column,xtype) : list(uniq(sorted(gcol(column)))) if xtype==str else xtype, gpairs(flist,xtypes)) # possible values (not necessary for continuous features)
				tree_features = map(lambda (index,values) : ("f%i"%(index+1), values), gpairs(flist,xvalues)) # create (name, values) pairs for tree.Learner constructor
				sublearn = frac.tree.Learner(algo="J48" if target_type==str else "REPTree",features=tree_features, dir=cachesub, clean=False ) 

				# make train/testset predictions
				sys.stderr.write("# Learn Tree predictor for feature %i of %i (%s)\n" % (target, N, time.strftime("%Y %b %d %H:%M:%S")))
				(trainset, testset) = ( map(lset,A) , map(lset,B) )
				f = sublearn( filter(lambda (x,y) : y is not None, trainset ));
				yield ( target, target_type, \
						gmap(lambda (y,g):(y,g), frac.trainset_predict(trainset, sublearn, N=folds)) ,\
						gpairs(gmap(lambda (x,y) : y, testset), f.predictions(map(lambda (x,y) : x, testset))) )

# Map training,testing predictions to a measure of surprisal per feature
# Return (Svals, Spred, entropy):
#	surprisal (for each test examle) given trainset distribution only, 
#	surprisal given predictors, and 
# 	the entropy of the training values.
def surprisal( Gtrain, Gtest, ftype ):

	Gtrain=list(Gtrain);
	Gtest = list(Gtest);

	if ftype in [str]: # nominal feature
		
		# Note: I'm not looking at the test set feature distribution; I'm just
		# collecting all possible feature values, which are part of the
		# definition of the nominal feature type 
		possible_values = list(uniq(gfilter(lambda y : y is not None, gmap(lambda (y,g):y, gcat([Gtrain,Gtest]))))) # all known feature values ('None' may be present, but we already know that represents a missing data point)

		valmap = dict(map(lambda (i,val) : (val,i), iwalk( sorted(possible_values) )))
		confusion = map(lambda _ : multinomial.Multinomial([1]*len(valmap)), range(len(valmap)))	# starting with pseudocount of 1
		valuedist = multinomial.Multinomial([1]*len(valmap)) # starting with pseudocount

		for (y,g) in filter(lambda (y,g) : y is not None, Gtrain):
			confusion[valmap[g]].d[valmap[y]] += 1	# confusion: one multinomial for each prediction
			valuedist.d[valmap[y]] += 1
		for m in confusion:	
			m.normalize()
		valuedist.normalize()
		entropy = multinomial.entropy( valuedist.d )
		
		Spred = map(lambda (y,g) : None if y is None else multinomial.surprisal( confusion[valmap[g]].d[valmap[y]] ), Gtest)
		Svals = map(lambda (y,g) : None if y is None else multinomial.surprisal( valuedist.d[valmap[y]] ), Gtest)

		## Informative statistics on feature value distributions with and
		## without the predictor (not required for method):
		preddist = multinomial.Multinomial([0]*len(valmap))
		for (y,g) in Gtrain:
			preddist.d[valmap[g]] += 1
		preddist.normalize()
		sys.stderr.write("# Without model: %s (entropy=%g)\n# With model:    %s (entropy=%g)\n# Information Gain = %s\n" % (\
			', '.join(map(lambda (val,index) : "p(%s)=%g" % (val,valuedist.d[index]),  valmap.iteritems() )),\
			entropy, \
			"; ".join( map(lambda (pval,pindex) : ", ".join(map( lambda (yval,yindex) : "p(%s|%s)=%g" % (yval,pval,confusion[pindex].d[yindex]), valmap.iteritems())) , valmap.iteritems() )), \
			sum(map( lambda (val,index) : preddist.d[index] * multinomial.entropy(confusion[index].d), valmap.iteritems() )), \
			entropy - sum(map( lambda (val,index) : preddist.d[index] * multinomial.entropy(confusion[index].d), valmap.iteritems() )) ))
		
	else: # continuous feature

		'''
		## METHOD 1:  Forget method 1
		'''

		'''
		## METHOD 4:  Represent PDF with sum of Gaussians, subtract entropy
		epsilon = 1e-9;
		Y = map(lambda (y,g) : y, filter(lambda (y,g) : y is not None, Gtrain)) # Y <=> training set feature values
		minY,maxY = min(Y)-epsilon, max(Y)+epsilon
		ydist = sog.SumOfGaussians(Y, stats.stdev(Y) + epsilon) # model of feature values
		entropy = 0 if minY==maxY else ydist.entropy( minY, maxY ); 	# entropy of training feature values

		E = map(lambda (y,g) : y-g, filter(lambda (y,g) : y is not None, Gtrain)) # E <=> training set error distribution
		minE,maxE = min(E)-epsilon,max(E)+epsilon
		edist = sog.SumOfGaussians(E, stats.stdev(E) + epsilon) 	# let "edist" be the error distribution over the trainset

		Svals = map(lambda (y,g) : None if y is None else ydist.surprisal(  y , min(minY, y ), max(maxY, y ) ), Gtest) # surprisal of predictions
		Spred = map(lambda (y,g) : None if y is None else edist.surprisal( y-g, min(minE,y-g), max(maxE,y-g) ), Gtest) # surprisal of error

		## Informative statistics on feature value distributions with and without the predictor (not required for method)
		pent = 0 if minE==maxE else edist.entropy(minE,maxE)
		sys.stderr.write("# Without model: [%s,%s], %s +/- %s (entropy = %s)\n# With model:    [%s,%s], %s +/- %s (entropy = %s)\n# Information Gain = %s\n" % (\
			minY,maxY,stats.mean(Y),stats.stdev(Y),entropy, \
			minE,maxE,stats.mean(E),stats.stdev(E),pent, \
			entropy - pent))
		'''


		## METHOD 5:  Smoothed Histogram
		n = len(filter(lambda (y,g):y is not None,Gtrain)) # number of [valid] training instances
		bins = max(2, int(math.sqrt(n)))
		sigma = bins/math.sqrt(n);

		class Histogram: # maps values to bin space
			def __init__(self, D, bins, sigma, epsilon=1e-9):	
				D = list(D)
				self.bins, self.sigma, self.minv, self.maxv, self.epsilon = bins, sigma, min(D), max(D), epsilon
				self.bins *= 3; 	# expand domain in case of testset under/overflow
				domain = self.maxv - self.minv;
				self.maxv += domain + epsilon
				self.minv -= domain + epsilon
				self.H = [0]*self.bins
				for x in D:	
					self.H[ self.map(x) ] += 1
				self.H = map(lambda count : count/float(len(D)), self.H) # normalize
				self.H = multinomial.smooth(self.H, sigma = sigma)
			def map(self, x):
				#return int(self.bins * (x - self.minv) / (self.maxv - self.minv)) # x maps to this bin
				bin = self.bins * (x - self.minv) / (self.maxv - self.minv) # x maps to this bin
				return self.bins-1 if bin==len(self.H) else int(bin)
			def __call__(self, x):
				bin = self.map(x)
				return max(self.H[bin],self.epsilon) if 0 <= bin < len(self.H) else self.epsilon # return tiny epsilon if probability goes to zero or if out of range

		E = Histogram( map(lambda (y,g):y-g, gfilter(lambda (y,g):y is not None,Gtrain)), bins, sigma )
		Y = Histogram( map(lambda (y,g):y  , gfilter(lambda (y,g):y is not None,Gtrain)), bins, sigma )

		entropy = multinomial.entropy(Y.H)	# Entropy of the training values
		Spred = map(lambda (y,g) : None if y is None else multinomial.surprisal(E(y-g)), Gtest)
		Svals = map(lambda (y,g) : None if y is None else multinomial.surprisal(Y( y )), Gtest)
	
		# Optional:  Say some nice things about the results
		YD = map(lambda (y,g): y , gfilter(lambda (y,g):y is not None,Gtrain))
		ED = map(lambda (y,g):y-g, gfilter(lambda (y,g):y is not None,Gtrain))
		sys.stderr.write("# Without model: [%s,%s], %s +/- %s (entropy of smoothed histogram = %s)\n# With model:    [%s,%s], %s +/- %s (entropy of smoothed histogram = %s)\n# Information Gain (smoothed histogram) = %s\n" % ( min(YD),max(YD),stats.mean(YD),stats.stdev(YD),multinomial.entropy(Y.H), min(ED),max(ED),stats.mean(ED),stats.stdev(ED),multinomial.entropy(E.H), multinomial.entropy(Y.H) - multinomial.entropy(E.H))) 
		#sys.stderr.write("# Feature value [%s,%s]:\n\n# %s\n\n%s\n\n" % (Y.minv,Y.maxv,sorted(YD),'\n'.join(map(lambda (i,d):"%s\t%s\n%s\t%s"%( Y.minv+i*(Y.maxv-Y.minv)/Y.bins,d,Y.minv+(i+1)*(Y.maxv-Y.minv)/Y.bins,d),iwalk(Y.H)))))
		#sys.stderr.write("# Error dist: [%s,%s]\n\n%s\n\n" % (E.minv,E.maxv,'\n'.join(map(lambda (i,d):"%s\t%s\n%s\t%s"%( E.minv+i*(E.maxv-E.minv)/E.bins,d,E.minv+(i+1)*(E.maxv-E.minv)/E.bins,d),iwalk(E.H)))))

	return (Svals, Spred, entropy)	# (Surprisal of values, surprisal of values | predictions, entropy of values)

# Create files with all sorts of predictions
def suite( base, trainsz, testsz ):

	FEATURE_COL = 0
	TYPE_COL = 1
	BEGIN_PAIRS_COL = 2

	# these are all running totals parallel to the testset examples
	NUM_STATS = 4
	#SURPRISAL,NORMALIZED_SURPRISAL,TOTAL_ERROR,NORMDIST_SURPRISAL,CONSTRAINED,VOTING,SURPRISAL_RATIO = range(NUM_STATS)
	SURPRISAL,NORMALIZED_SURPRISAL,TOTAL_ERROR,NORMDIST_SURPRISAL = range(NUM_STATS);
	stats = map(lambda _ : map(lambda _ : 0, range(testsz)), range(NUM_STATS))

	basein = open(base,'r');
	for record in map(lambda line:line.split(), filter(lambda line:line,map(lambda line:line.split('#',1)[0], basein))):	# for each feature

		ftype = eval(record[TYPE_COL]);
		resolve = lambda symbol : None if symbol==str(None) else ftype(symbol);
		pairs = record[BEGIN_PAIRS_COL:];
		Gtrain = map(lambda (y,g) : (resolve(y),resolve(g)), map(lambda p : p.split(','), pairs[:trainsz]))
		Gtest  = map(lambda (y,g) : (resolve(y),resolve(g)), map(lambda p : p.split(','), pairs[trainsz:]))
		
		assert len(Gtrain) == trainsz and len(Gtest) == testsz;	# fails when record length is wrong
		#sys.stderr.write("# (feature %i) READ IN\nGTRAIN:%s\nGTEST:%s\n" % (int(record[FEATURE_COL]),Gtrain,Gtest));
		
		sys.stderr.write("# --\n# Feature %s (%s, %s training, %s testing) %s\n" % (record[FEATURE_COL], record[TYPE_COL], trainsz, testsz, time.strftime("%Y %b %d %H:%M:%S")))
		(Svals,Spred,entropy) = surprisal(Gtrain,Gtest,ftype)

		# add value based on this feature to running totals for each test example
		for (i, (Sv, Sp)) in filter(lambda (i,(Sv,Sp)):Sv is not None or Sp is not None, iwalk(gpairs(Svals,Spred))):
			stats[SURPRISAL][i] += Sp;	# surprisal is ZERO when feature value is missing
			stats[NORMALIZED_SURPRISAL][i] += Sp - entropy; 
			stats[NORMDIST_SURPRISAL][i] += Sv;
			#stats[CONSTRAINED][i] += max(-1,min(post-prior,1))	# zero when feature missing
			#stats[VOTING][i] += 1 if post>prior else -1 # and abstain when feature is missing
			#stats[SURPRISAL_RATIO][i] += 1 if prior==0 else post/prior # define undefined ratio as 1; fine

		for (i,(y,g)) in filter(lambda (i,(y,g)):y is not None, filter(lambda (i,G):G is not None,iwalk(Gtest))):
			stats[TOTAL_ERROR][i] += (0 if y==g else 1) if ftype in [str] else (abs(y-g))	 # error is ZERO when observed feature value is missing

	write_fun = lambda ext,ag : open("%s.%s" % (base,ext),'w').write(''.join(gmap(lambda x : "%s\n" % x, ag)))
	write_fun('error', stats[TOTAL_ERROR]);
	write_fun('surprisal', stats[SURPRISAL])
	write_fun('normalized-surprisal', stats[NORMALIZED_SURPRISAL])
	write_fun('normdist-surprisal', stats[NORMDIST_SURPRISAL])
	#write_fun('constrained', stats[CONSTRAINED])
	#write_fun('voting', stats[VOTING])
	#write_fun('surprisal-ratio', stats[SURPRISAL_RATIO])
		

# reformat a train/test set of pairs:  Numericize predictor features.
def svm_reformat(A, B):
	X = map(lambda (x,y) : x, A+B)	# predictors
	Y = map(lambda (x,y) : y, A+B)	# targets
	XY = list(gpairs( format.numericize(X), Y ))
	return (XY[:len(A)], XY[len(A):])

def chkdir(path):
	if not os.path.exists(path):
		os.makedirs(path)
	return path
if (__name__ == "__main__"):
	main();
