/*

	2010/Mar, noto@cs.tufts.edu

	This is a modified version of LOF (from Umaa) which takes TRAINING and TEST
	(see comments for constants N,C, and C0) examples, and considers each TEST
	example to NOT be part of the general population (e.g. a TEST example may
	still be an outlier if it is in the same neighborhood with other TEST
	examples), and calculates the TEST set LOF scores en masse.

	Input is comma-[and newline]-separated data.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

// distance(i,j) stored in `stored_dist' (in main).  Only allocated stored_dist where i > j 
#define dist(i,j) ((i > j) ? stored_dist[i][j] : stored_dist[j][i])

double distance(const double *x, const double *y, const int dim);

/** Used to measure each object's neighbors (need distance for sorting) */
typedef struct {
	int index; 
	double distance;
} neighbor;

// for sorting neighbors ascending by distance
int neighborcmp(const void *a, const void *b);

int main(const int argc, const char **argv) {

	FILE *verbose = stderr; 	// set to NULL to turn off verbose progress messages.
	int vperiod;	// inside loops, print progress every vperiod cycles.
	time_t t0 = time(NULL);

	int i,j,p,k,f;

	if (argc != 7) {
		fprintf(stderr,"ERROR: expecting command line arguments: \n");
		fprintf(stderr,"MinPts lower bound, MinPts upper bound, no. training instances, no. testing instances, no. features, input file name\n");
		exit(-1);
	}

	const int T = atoi(argv[3]);		// training instances
	const int M = T + atoi(argv[4]);	// total instances
	const int F = atoi(argv[5]);		// num features

	int user_MinPtsLB = atoi(argv[1]); 
	if (user_MinPtsLB < 2) { fprintf(stderr,"Setting MinPts lower bound <- 2 (%i is too small)\n", user_MinPtsLB); user_MinPtsLB=2; }
	if (user_MinPtsLB >= T) { fprintf(stderr,"Setting MinPts lower bound <- %i (%i is too large)\n", T-1, user_MinPtsLB); user_MinPtsLB=T-1; }
	int user_MinPtsUB = atoi(argv[2]); 
	if (user_MinPtsUB < 2) { fprintf(stderr,"Setting MinPts upper bound <- 2 (%i is too small)\n", user_MinPtsUB); user_MinPtsUB=2; }
	if (user_MinPtsUB >= T) { fprintf(stderr,"Setting MinPts upper bound <- %i (%i is too large)\n", T-1, user_MinPtsUB); user_MinPtsUB=T-1; }
	
	const int MinPtsLB = user_MinPtsLB;
	const int MinPtsUB = user_MinPtsUB;

	// The following change standard LOF to the special version we use to
	// estimate LOF over a train and test set.  The basic idea is that test
	// examples are "invisible" to everybody else so that each test can be
	// calculated independently.

	// To make this 'standard' LOF, which calculates a score for all input
	// examples, set C0 = 0, C = N = M.

	const int N = T;	// number of available neighbors (set to T to ignore test examples as neighbors)
	const int C = M-T;	// number of examples for which to calculate LOF
	const int C0 = T;	// 1st example for which to calculate LOF

	if (MinPtsLB <= 0) { fprintf(stderr,"ERROR: k lower bound must be greater than 0: %s\n", argv[1]); }
	if (M <= 0) { fprintf(stderr,"ERROR: M must be greater than 0: %s\n", argv[4]); }
	if (F <= 0) { fprintf(stderr,"ERROR: F must be greater than 0: %s\n", argv[5]); }

	vperiod = T/20 + 100 - (T/20)%100;	// adjust to a percentage of training set, round to nearest 100

	// allocate space for data (examples x features)
	double **data = (double**)malloc(sizeof(double*) * M);	// freed after stored_dist is calculated
	for (i=0; i<M; i++) { data[i] = (double*)malloc(sizeof(double)*F); }

	// space for pairwise distance (need only lower triangle of training x
	// training instances, and distance from test instances to training
	// instances)
	double **stored_dist = (double**)malloc(sizeof(double*) * M); 	// freed at the end
	for (i=0; i<T; i++) { stored_dist[i] = (double*)malloc(sizeof(double)*i); } // must be i > j
	for (i=T; i<M; i++) { stored_dist[i] = (double*)malloc(sizeof(double)* (N < i ? N : i)); }
	
	// read data
	if (verbose) { fprintf(verbose, "# Read input...\n"); }
	FILE *in = fopen(argv[6],"r");
	for (i=0; i<M; i++) { 
		for (f=0; f<F; f++) { 
			fscanf(in, "%lf,", data[i]+f); 
		}
	}
	fclose(in);

	
	t0 = time(NULL);
	if (verbose) { fprintf(verbose, "# Compute pairwise distance for %i training examples",  T); }
	for (i=0; i<T; i++) { 
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
		for (j=0; j<i; j++) { 
			stored_dist[i][j] = distance( data[i], data[j], F );
		}
	}
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Compute pairwise distance for %i test examples", difftime(time(NULL),t0), M-T); }
	t0 = time(NULL);

	for (i=T; i<M; i++) {
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
		for (j=0; j<N; j++) {
			stored_dist[i][j] = distance( data[i], data[j], F ); 
		} 
	}
	
	// free original input data
	for (i=0; i<M; i++) { free(data[i]); }
	free(data);
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Find %i neighbors for each of %i examples", difftime(time(NULL),t0), MinPtsUB, M); }
	t0 = time(NULL);

	// The efficient way to proceed is to find the MinPtsUB (upper bound)
	// nearest neighbors for each example, then choose a subset among them for
	// all k <= MinPtsUB
		
	neighbor **knn = malloc(sizeof(neighbor*) * M); // a neighbor is an index and its distance, freed at the bottom of the program
	for (i=0;i<M;i++) { 
		knn[i] = malloc(sizeof(neighbor)*MinPtsUB); 
		for (k=0; k<MinPtsUB; k++) { 
			knn[i][k].index = -1;		
			knn[i][k].distance = 1e999;
		}
	}
	
	// loop to find nearest neighbors and calculate k-distance
  	for (i=0; i<M; ++i) {
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
		for (j=0; j<N; ++j) {
			if (i==j) { continue; }
	  		if (dist(i,j) < knn[i][MinPtsUB-1].distance) { 
				// we found a nearer neighbor

				// overwrite most distant neighbor
				knn[i][MinPtsUB-1].index = j;
				knn[i][MinPtsUB-1].distance = dist(i,j);

				// sort so that knn[i] is a list of MinPtsUB neighbors, closest one first, 
				// index knn[i][MinPts-1] is the most distant neighbor
				qsort(knn[i], MinPtsUB, sizeof(neighbor), neighborcmp);

	  		}
		}

	} // next i
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Compute LOF for %i <= k <= %i", difftime(time(NULL),t0), MinPtsLB, MinPtsUB); }

	// store the (max) LOF for C test examples
	double max_LOF[C];
	for (i=0; i < C; i++) { max_LOF[i]=0.0; }

	t0 = time(NULL);
	for (k=MinPtsLB; k<=MinPtsUB; k++) {

		if (verbose) { fprintf(verbose,"."); }
			
		double **reach_dist = malloc(sizeof(double*)*M);
		for (i=0; i<M; i++) { reach_dist[i] = malloc(sizeof(double)*k); } // freed at bottom of loop

	  	// reach-dist
	  	for (i=0; i<M; ++i) {
			for (j=0; j<k; ++j) {
		 		int nn_i = knn[i][j].index;
		  		reach_dist[i][j] = (dist(i,nn_i)) > knn[nn_i][k-1].distance ? (dist(i,nn_i)) : knn[nn_i][k-1].distance;
			}
	  	}


	  	// lrd
	  	double *lrd = malloc(sizeof(double)*M); 	// freed at the bottom of loop

	  	for (i=0; i<M; ++i) {
			double sum_reach_dist = 0.0;
			for (j=0; j<k; ++j) {
				sum_reach_dist += reach_dist[i][j];
			}
			lrd[i] = k / sum_reach_dist;

		}

		// LOF
		for (i=C0; i < C0+C; i++) { 

			//calculate LOF
			double sum_lrd = 0.0;
			for (p=0; p < k; ++p) {
		  		sum_lrd += lrd[knn[i][p].index]/lrd[i];
			}

			int test_index = i - C0;
			double LOF_k_test_index = sum_lrd / k;
			max_LOF[test_index] = (!isinf(LOF_k_test_index) && LOF_k_test_index > max_LOF[test_index]) ? LOF_k_test_index : max_LOF[test_index];	// keep max LOF
        
	  	}

		free(lrd);
		for (i=0; i<M; i++) { free(reach_dist[i]); }
		free(reach_dist);
	
	
	} // next value of k
		
	if (verbose) { fprintf(verbose, "%.0f second(s).\n", difftime(time(NULL),t0)); }
			
	for (i=0;i<M;i++) { free(knn[i]); }
	free(knn);

	// output:  print max(LOF(i)) for all test cases
	for (i=0; i<C; i++) { 
		printf("%g\n", max_LOF[i]);
	}

	// noto:  free stored_dist
	for (i=0; i<M; i++) { free(stored_dist[i]); }
	free(stored_dist);

	return 0;
}


double distance(const double *x, const double *y, const int dim) { 
	int i;
	double ssd = 0.0;
	for (i=0; i<dim; i++) { 
		double diff = x[i] - y[i];
		ssd += diff * diff;
	}
	return sqrt(ssd);
}

int neighborcmp(const void *a, const void *b) { 
	neighbor *na = (neighbor*)a;
	neighbor *nb = (neighbor*)b;
	if (na->distance < nb->distance) { return -1; }
	if (na->distance > nb->distance) { return  1; }
	if (na->index    < nb->index)    { return -1; }
	if (na->index    > nb->index)    { return  1; }
	return 0;
}



