#!/usr/bin/env python
# -*- coding: utf8 -*-

'''

Utilities related to probability density functions

The PDF objects represent probability density functions.  Each object must
implement the `p' and `sample' methods.

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.0";

import sys,string,random,math;

# PDF: A probability density function object
class PDF:

	# Calculate probability density p(x)
	def p(self,x):
		pass; 

	# Sample from the PDF.  Return a real number
	def sample(self):
		pass;

	# Calcualte cumulative density, i.e. SUM[-inf,x].   
	# "Generic" PDF version adds tiny slices
	def CDF(self,x):
		gran = 0.001; # granularity
		p1 = self.p(x);
		p0 = self.p(x - gran); 
		mass = slice = gran * ((p1 + p0) / 2.0);
		while (slice > 0.000001):	# last section was too small
			x -= gran;
			p1 = p0; 
			p0 = self.p(x);
			slice = gran * ((p1+p0)/2.0);
			mass += slice;
		return mass;

	# Calculate the point at which CDF=p.  "Generic" version does a type of binary search.
	def inverseCDF(self,p):
		lo = -1.0; 
		while ( self.CDF(lo) > p ):
			lo *= 2;
		hi = 1.0; 
		while (self.CDF(hi) < p):
			hi *= 2;
		return bisectCDF(self.CDF, lo, hi, p);

	# Calculate/Estimate the mean.  "Generic" version uses sample mean.
	def mean(g):
		sum = 0.0;
		kmax = 10000;
		for i in range(0,kmax):
			sum += g.sample();
		return sum / kmax;
	
	# Calculate/Estimate the variance.  "Generic" version uses sample variance.
	def variance(g):
		mean = g.mean();
		ssd = 0.0;
		kmax = 1000000;
		for i in range(0,kmax):
			x = g.sample();
			ssd += (x-mean) * (x-mean);
		return ssd / kmax

	# Calcuate/Estimate the median 
	def median(g):
		return g.inverseCDF(0.5);

	# Calculate/Estimate the mode.  "Generic" version uses simulated annealing
	def mode(g):
		L = g.inverseCDF(0.05);	#  Well off to the left
		R = g.inverseCDF(0.95);	#  Right
		range = R - L;			#  Width of the Middle 90%
		neighborhood = Gaussian(0.0, range / 10);	# i.e. let range/2 be 5 std. deviations.
		neighbor = lambda s : s + neighborhood.sample();	
		state0 = g.inverseCDF(0.5);	# start in the middle 
		E = lambda s : -g.p(s); 	# lower is better
		kmax = 10000;				# number of potential moves
		temp = lambda elapsed : 1-elapsed;	# temperature goes from 1 to 0
		P = lambda e0,e1,T : (1 if (e1<e0) else (1 if (e0==0.0) else (T*(e1/e0))));		# P( move ) = "badness" * temperature
		return pdfsa( state0, E, neighbor, P, temp, kmax);

	# Print out a good look
	def dump(g, step = 0.01, out=sys.stdout):
		assert 0 < step < 1
		p = step;
		while p < 1:
			x = g.inverseCDF(p)
			out.write("%s\t%s\n" % ( x, g.p(x) ))
			p += step

# A Uniform density, i.e. all values [min,max] are equiprobable.
class Uniform(PDF):
	def __init__(self,min,max):
		self.min=min;
		self.max=max;
	def __str__(self):
		return "Uniform PDF [%s,%s]" % (self.min,self.max);
	def p(u,x):
		return 1.0 if (u.min <= x < u.max) else 0.0;
	def sample(u):
		return u.min + random.random() * (u.max - u.min);
	def mean(u):
		return (u.max+u.min)/2.0;
	def median(u):
		return (u.max+u.min)/2.0;
	def mode(u):
		return u.min;
	def CDF(u,x):
		return ( (x-u.min) / (u.max-u.min)) if (u.min <= x < u.max) else 0.0 if x < u.min else 1.0;
	def inverseCDF(u,p):
		return u.min + ((u.max - u.min) * p);
	def variance(u):
		b = (u.max - u.min)/2.0	# b difference between max and mean
		return b*b/3 # calculus!
		

class Gaussian(PDF):
	def __init__(self,mu=0.0,sigma=1.0):
		self.mu = float(mu);
		self.sigma = float(sigma);
		self.samples = [];	# see Gaussian.sample2
	def mean(g):
		return g.mu;
	def median(g):
		return g.mu;
	def mode(g):
		return g.mu;
	def variance(g):
		return g.sigma * g.sigma;
	
	def p(g,x):
		# 1 / sqrt(2PI) = 0.3989422804014327 
		d = x - g.mu;
		return (1/g.sigma) * 0.3989422804014327 * math.exp( -d*d / (g.sigma*g.sigma) ); 

	def CDF(g,x):
		z = (x - g.mu) / g.sigma;
		v = 0.5 * gammp( 0.5, z*z/2.0 );
		return -1.0 if (v<0) else -v+0.5 if (z<0) else v+0.5; 

	def inverseCDF(g,p): 
		return ( g.mu + g.sigma * standInverse(p) );

	# Sample two points from distribution
	def sample2(g):
		# http://en.wikipedia.org/wiki/Box-Muller_transform
		r = 0.0;
		(x,y) = (None,None);
		while (r==0.0 or r > 1.0):
			# generate two uniform randoms [1,1]
			x = random.random() * 2 - 1;
			y = random.random() * 2 - 1;
			r = (x*x) + (y*y);
		W = 0 if r==0.0 else math.sqrt( (-2*math.log(r)) / r );
		sample1 = (x * W) * g.sigma + g.mu;	
		sample2 = (y * W) * g.sigma + g.mu; 
		return (sample1, sample2);

	def sample(g):
		if (g.samples):
			ans = g.samples[0]; 
			g.samples = g.samples[1:];
			return ans;
		else:
			(x1,x2) = g.sample2(); 
			g.samples += [ x2 ];
			return x1;

	def __str__(g):
		return "Gaussian(mu=%s, sigma=%s)" % (g.mu,g.sigma);

# A weighted mixture of PDFs.
class Mixture(PDF):
	def __init__(self, mixture={}):
		self.mixture=mixture;	# { component:weight }
	def __str__(m):
		return "{%s}" % ('; '.join(map(lambda (c,w) : "%s, weight=%s" % (str(c),w), m.mixture.iteritems())))
	def p(m,x):
		return sum(map(lambda (c,w) : w*c.p(x), m.mixture.iteritems())); 
	def sample(m):
		return (max(map( lambda (c,w) : (w*random.random(),c) , m.mixture.iteritems() ))[1]).sample();
	def CDF(m,x):
		return sum(map(lambda (c,w) : w*c.CDF(x), m.mixture.iteritems())); 
	def mean(m):
		# mean is always weighted mean (?)
		sum = 0.0;
		mass = 0.0;
		for (c,weight) in m.mixture.iteritems():
			sum += weight * c.mean();
			mass += weight;	# I'm mixing my metaphors
		return sum / mass;	# mass should be 1
		

class Gamma(PDF):
	def __init__(self,alpha=2,beta=1):
		self.alpha=float(alpha);
		self.beta=float(beta);
	def mean(g):
		return g.alpha * g.beta;
	def mode(g):
		if (g.alpha >= 1.0):
			return (g.alpha - 1)*g.beta;
		else:
			return PDF.mode(g);
	def variance(g):
		return g.alpha *g.beta * g.beta;
	def __str__(g):
		return "Gamma(alpha=%s, beta=%s)" % (g.alpha,g.beta);
	def p(g,x):
		if (x < DELTA):
			fval = 0.0;
		else:
			if (g.alpha < LANCZ_CUTOFF):
				fval = lanczos(g.alpha);
			else:
				fval = logStirling(g.alpha);
			fval = math.exp( g.alpha * math.log(1/g.beta) - fval + (g.alpha - 1.0) * math.log(x) - (1/g.beta) * x );
		return fval;
	def CDF(g,x):
		if (x <= 0):
			return 0;
		else:
			return (gammp(g.alpha, x*(1/g.beta)));
	def sample(g):
		# Fishman, "Sampling from the Gamma Distribution on a Computer",
		# Management Science/Operations Research, 1976, Association for
		# Computing Machinery, Inc.
		a = g.alpha - 1;
		while (1): 
			V = -math.log( random.random() );
			W = -math.log( random.random() );
			if (W >= a*(V-math.log(V)-1)): 
				return g.alpha * V * g.beta; 


ITMAX        =     300;
FPMIN        = 1.0E-50;
EPSILON      = 1.0E-20;
DELTA        = 1.0E-13;
LANCZ_CUTOFF =   700.0;
MAXPOintS=    8000;
DFLT_AXIS_RATIO=  0.93;
DFLT_TOP_RATIO =  0.05;
DFLT_BOT_RATIO =  0.12;
MAX_BOT_RATIO  =  0.19;

def logBinCoef(n,k):

	if ( n >= LANCZ_CUTOFF ):
		Lbc = 1.0 - 0.5 * math.log(2.0 * PI);
		Lbc += (n + 1.5) * math.log(n + 1.0);
		Lbc -= (k + 1.5) * math.log(k + 1.0);
		Lbc -= (n + 1.5 - k) * math.log(n - k + 1.0);
	else:
		Lbc = lanczos(n + 1.0) - lanczos(k + 1.0) - lanczos(n - k + 1.0);
	return Lbc;

def logStirling(p):
	return 0.5 * math.log(2.0 * math.pi) + (p+0.5) * math.log(p) - p;


# bisectCDF() find the inverse of CDF() using bisection
def bisectCDF(CDF, x_lo, x_hi, prob):

	k=0;
	Iters = 60;  # assuring 18 digits of decimal precision
	hi = x_hi;
	lo = x_lo;
	mid = 0.5 * (hi + lo);

	# 1) determine whether the CDF is computable at all 
	if (CDF(mid) < 0.0):
		return (x_lo);

	# 2) extend the right bisection window endpoint if necessary 
	k = 0;
	while (k==0 or k < ITMAX):
		cdf = CDF(hi);
		if ((cdf >= prob) or (cdf < 0.0)):
			break;
		hi += 5.0;
		k += 1;

	if (k > 0): 
		lo = x_hi;
	else:
		# 3) extend the left end of the bisection window if necessary 
		k = 0;
		while (k==0 or k < ITMAX):
			cdf = CDF(lo);
			if (cdf <= prob):
				break;
			lo -= 5.0;
		if (k > 0):
			hi = x_lo;

	# 4) bisect the puppy 
	k = 0;
	while (k==0 or ((hi-lo > EPSILON) and (k < Iters))):
		mid = 0.5 * (hi + lo);
		cdf = CDF(mid);
		if (cdf < 0): # error
			mid = x_lo;
			break;
		if (abs(cdf - prob) < EPSILON):
			break;
		if (cdf > prob): 
			hi = mid;
		else:
			lo = mid;
		k += 1;
	
	return (mid);



'''
*
*  The following recipes are transcribed from "Numerical Recipes in C", 
*  Second Edition, by Press, Teukolsky, Vetterling, and Flannery, 
*  Cambridge University Press, 1992.
*
*
'''
def betai(a,b,x):

	if (x < 0.0 or x > 1.0):
		return -1.0;
	elif (x == 0.0 or x == 1.0):
		return 0;
	else:
		bt = math.exp( lanczos(a+b) - lanczos(a) - lanczos(b) + a*math.log(x) + b*math.log(1.0-x) );
		if (x < ((a + 1.0) / (a + b + 2.0))):
			cf = betacf(a,b,x); 
			if (cf is None):
				return -1.0;
			return bt * cf / a; 
		else:
			cf = betacf(b, a, 1.0 - x);
			if (cf is None):
				return -1.0;
			return 1.0 - bt * cf / b;


# Continued fraction for incomplete beta function by modified Lentz's method
def betacf(a, b, x):
	qab = a + b;
	qap = a + 1.0;
	qam = a - 1.0;
	c = 1.0;
	d = 1.0 - qab * x / qap;
	if (abs(d) < FPMIN):
		d = FPMIN;
	d = 1.0 / d;
	retval = d;
	for m in range(1,ITMAX+1):
		m2 = m + m;
		aa = (b-m) * m * x / ( (qam+m2) * (a+m2) );
		d = aa * d + 1.0;
		if (abs(d) < FPMIN):
			d = FPMIN;
		c = aa / c + 1.0;
		if (abs(c) < FPMIN):
			c = FPMIN;
		d = 1.0 / d;
		retval *= d * c;
		aa = (-(a+m) * (qab+m) * x) / ( (a+m2) * (qap+m2) );
		d = aa * d + 1.0;
		if (abs(d) < FPMIN):
			d = FPMIN;
		c = aa / c + 1.0;
		if (abs(c) < FPMIN):
			c = FPMIN;
		d = 1.0 / d;
		delv = d * c;
		retval *= delv;
		if (abs(delv - 1.0) < EPSILON):
			break;
	if (m > ITMAX):
		return None;	# bad value
	return retval;


# Lanczos's neat approximation of math.log(Gamma(p))
def lanczos(p):
	x = p;
	tmp = x + 5.5;
	tmp = tmp - (x + 0.5 ) * math.log(tmp);
	ser = 1.000000000190015 + 76.18009172947146 / ( p + 1 );
	ser -= 86.50532032941678 / ( p + 2 );
	ser += 24.01409824083091 / ( p + 3 );
	ser -= 1.231739572450155 / ( p + 4 );
	ser += 0.001208650973866179 / ( p + 5 );
	ser -= 0.000005395239384953 / ( p + 6 );
	return (math.log(2.506628274631001 * ser / x) - tmp);


def gammp(a,x):
	if (x < 0 or a <= 0):
		retval = -1.0;
	else:
		if (a >= LANCZ_CUTOFF):
			g = logStirling(a);
		else:
			g = lanczos(a);
	if (x < a + 1.0):
		retval = gser(a, x, g);
	else:
		g = gcf(a, x, g);
		if (g < 0):
			retval = g;
		else:
			retval = 1.0 - g;
	return retval;


def gammq(a, x):
	if (x < 0 or a <= 0):
		retval = -1.0;
	else:
		if (a >= LANCZ_CUTOFF):
			g = logStirling(a);
		else:
			g = lanczos(a);
		if (x >= a + 1.0): 
			retval = gcf(a, x, g);
		else:
			g = gser(a, x, g);
			if (g < 0):
				retval = g;
			else: 
				retval = 1.0 - g;
	return retval;


def gcf(a,x,g):
	b = x + 1.0 - a;
	c = 1.0 / FPMIN;
	d = 1.0 / b;
	h = d;
	for i in range(1, ITMAX+1):
		an = i * (a - i);
		b += 2.0;
		d = an * d + b;
		if (abs(d) < FPMIN):
			d = FPMIN;
		c = b + an / c;
		if (abs(c) < FPMIN):
			c = FPMIN;
		d = 1.0 / d;
		delv = d * c;
		h *= delv;
		if (abs(delv - 1.0) < EPSILON):
			break;
	if (i > ITMAX):
		retval = -1.0;
	else:
		retval = math.exp(a * math.log(x) - x - g) * h;
	return (retval);


def gser(a, x, g):
	retval=-1.0;
	if (x == 0.0):
		retval = 0.0;
	elif (x > 0):
		ap = a;
		sum = 1.0 / a;
		delv = sum;
		for i in range(1,ITMAX+1):
			ap += 1.0;
			delv *= x / ap;
			sum += delv;
			if (abs(delv) < abs(sum) * EPSILON):
				retval = sum * math.exp(a * math.log(x) - x - g);
				break;
	return (retval);


def standInverse(area):
	k=0;
	z=0.1;
	CDF = abs(area - 0.5) + 0.5;
	while (k==0 or ((abs(f - CDF) > DELTA) and (ITMAX > k))):
		p = 0.5 * gammp( 0.5, z * z / 2.0 );
		f = (-p+0.5) if (z<0) else p+0.5;
		z -= math.exp(.5 * z * z) * (f - CDF);
		k += 1;
	if (area < .5):
		z = 0.0 - z;
	return (z);


# Resident simulated annealing implementation
# @param state0 initial state
# @param E energy function:  map state to energy (lower is better)
# @param neighbor neighbor function:  map state to next state
# @param P probability function:  map old energy, new energy, temperature to a probability
# @param temp temperature schedule:  map ratio of elapsed time [0,1] to temperature
# @param kmax maximum time steps
# @param emax acceptable energy 
def pdfsa(state0, E, neighbor, P, temp, kmax, emax=-10e99999):
	(s,e) = (state0, E(state0));
	if (e < emax):
		return s;
	(sb, eb) = (s,e); 	# best
	for k in range(1, kmax+1):	# kmax time steps
		sn = neighbor(s);
		en = E(sn);
		if (en < eb):
			(sb,eb) = (sn,en);	# save new best
			if (eb < emax):
				break;
		if ( P(e,en,temp(float(k)/kmax)) > random.random() ):	# should we move?
			(s,e) = (sn,en);
	return sb;

# From The Endeavor:  http://www.johndcook.com/blog/2009/01/19/stand-alone-error-function-erf
# Here’s a Python implementation of erf(x) based on formula 7.1.26 from A&S. The maximum error is below 1.5 × 10-7.
# http://www.amazon.com/gp/product/0486612724?ie=UTF8&tag=theende-20&linkCode=xm2&camp=1789&creativeASIN=0486612724
def erf(x):
	# constants
	a1 =  0.254829592
	a2 = -0.284496736
	a3 =  1.421413741
	a4 = -1.453152027
	a5 =  1.061405429
	p  =  0.3275911

	# Save the sign of x
	sign = 1
	if x < 0:
		sign = -1
	x = abs(x)

	# A & S 7.1.26
	t = 1.0/(1.0 + p*x)
	y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*math.exp(-x*x)

	return sign*y

def test():
	
	import sys;
	import ag;	# eliminate a few lines below if this is unavailable

	sys.stderr.write("TEST MODULE:")
	sample_size = 100000

	models = [ Uniform(0,1), Gaussian(2,5), Gamma(2,1), Mixture( { Gaussian(-3,1):0.7, Gaussian(3,1):0.3 } ) ]
	
	for model in models:
		sys.stderr.write("\n--\n\n")
		sys.stderr.write("Model is %s\n" % model)
		sys.stderr.write("CDF(0) = %s\n" % model.CDF(0))
		sys.stderr.write("CDF(1) = %s\n" % model.CDF(1))
		sys.stderr.write("inverseCDF(0.01) = %s\n" % model.inverseCDF(0.01))
		sys.stderr.write("inverseCDF(0.5)  = %s\n" % model.inverseCDF(0.5))
		sys.stderr.write("inverseCDF(0.99) = %s\n" % model.inverseCDF(0.99))
		sys.stderr.write("CDF(inverseCDF(0.99)) = %s\n" % model.CDF(model.inverseCDF(0.99)))
		sample = map(lambda i : model.sample(), range(sample_size));
		sys.stderr.write("Mean = %s, sample mean = %s\n" % ( model.mean(), ag.mean(sample) ))
		sys.stderr.write("Median = %s, sample median = %s\n" % ( model.median(), ag.median(sample) ))
		sys.stderr.write("Variance = %s, sample variance = %s\n" % ( model.variance(), ag.variance(sample, False) ))
		sys.stderr.write("Mode = %s, sample mode (rounded to integers) = %s\n" % ( model.mode(), ag.mode(map(lambda x : math.floor(x), sample))))

	sys.stderr.write("Write graph of last model (%s) to stdout...\n" % (model))
	model.dump( step = 0.02 ,  out=sys.stdout )

if (__name__=='__main__'):
	test();


